"""Classes and functions used to handle data.

Note
----
New users may find the following tutorials helpful:

- `Loading Data <https://osl-dynamics.readthedocs.io/en/latest/tutorials_build\
  /data_loading.html>`_
- `Preparing Data <https://osl-dynamics.readthedocs.io/en/latest\
  /tutorials_build/data_preparation.html>`_
"""

from osl_dynamics.data.base import Data

__all__ = ["Data"]
